/*
    net.c

    This file is a part of Anubis.
    Copyright (C) 2001, 2002 The Anubis Team.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "headers.h"
#include "extern.h"

static int connect_directly_to(char *, unsigned int);
static int mread(int, void *, char *);

static int connect_directly_to(char *host, unsigned int port)
{
	int sd = 0;
	struct sockaddr_in addr;
	struct hostent *hp = 0;

	/****************************
	   Find out the IP address
	*****************************/

	info(VERBOSE, "Getting remote host information...");
	if ((hp = gethostbyname(host)) == 0) {
		addr.sin_addr.s_addr = inet_addr(host);
		if ((int)addr.sin_addr.s_addr == -1) {
			anubis_error("Unknown host %s", host);
			return -1;
		}
	}
	else
		memcpy((char *)&addr.sin_addr.s_addr, hp->h_addr, hp->h_length);

	/******************************
	   Create socket and connect
	*******************************/

	info(VERBOSE, "Initializing socket...");
	if ((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		anubis_error("Can't create stream socket.");
		return -1;
	}
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);

	#ifndef __BEOS__
	setsockopt(sd, SOL_SOCKET, SO_LINGER, 0, 0);
	setsockopt(sd, SOL_SOCKET, SO_KEEPALIVE, 0, 0);
	#endif /* not __BEOS__ */

	info(VERBOSE, "Connecting to %s:%u...", host, port);
	if (connect(sd, (struct sockaddr *)&addr, sizeof(addr)) == -1) {
		anubis_error("Couldn't connect to %s:%u", host, port);
		return -1;
	}
	else
		info(NORMAL, "Connected to %s:%u", host, port);

	return sd;
}

int make_remote_connection(char *host, unsigned int port)
{
	int sd;
	char host_backup[65];
	unsigned int port_backup;
	memset(host_backup, 0, sizeof(host_backup));

	/*
	   First of all, make a copy of 'host' and 'port'.
	*/

	safe_strcpy(host_backup, host);
	port_backup = port;

	check_all_proxies(host_backup, &port_backup);
	sd = connect_directly_to(host_backup, port_backup);
	if (sd == -1) /* an error */
		return -1;
	if (check_socks_proxy(sd, host, port) == -1)
		return -1;
	if (check_wingates(sd, host, port) == -1)
		return -1;

	return sd;
}

/***************************************
  Not exactly a network connection...
  Connect to stdin/stdout.
****************************************/

#ifndef NOFORK
static int make_sockets(int fd[2])
{
	#ifdef NOSOCKETPAIR
	struct sockaddr_in addr;
	int addrlen;
	int sd;

	if ((sd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		anubis_error("#1 socket() failed.");
		return -1;
	}
	if ((fd[1] = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		anubis_error("#2 socket() failed.");
		return -1;
	}
	addrlen = sizeof(addr);
	memset(&addr, 0, addrlen);
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	addr.sin_port = 0;
	if (bind(sd, (struct sockaddr *)&addr, addrlen)) {
		anubis_error("#1 bind() failed. %s", strerror(errno));
		return -1;
	}
	if (bind(fd[1], (struct sockaddr *)&addr, addrlen)) {
		anubis_error("#2 bind() failed. %s", strerror(errno));
		return -1;
	}
	if (listen(sd, 5)) {
		anubis_error("listen() failed. %s", strerror(errno));
		return -1;
	}
	if (getsockname(sd, (struct sockaddr *)&addr, &addrlen)) {
		anubis_error("getsockname() failed. %s", strerror(errno));
		return -1;
	}
	if (connect(fd[1], (struct sockaddr *)&addr, addrlen)) {
		anubis_error("connect() failed. %s", strerror(errno));
		return -1;
	}
	if ((fd[0] = accept(sd, (struct sockaddr *)&addr, &addrlen)) < 0) {
		anubis_error("accept() failed. %s", strerror(errno));
		return -1;
	}
	close_socket(sd);
	#else
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, fd)) {
		anubis_error("socketpair() failed. %s", strerror(errno));
		return -1;
	}
	#endif /* NOSOCKETPAIR */
	return 0;
}

static void sig_local(int code)
{
	pid_t pid;
	int status;

	while ((pid = waitpid(-1, &status, WNOHANG)) > 0)
		info(VERBOSE, "Local program [%d] finished.", pid);
	return;
}
#endif /* not NOFORK */

int make_local_connection(char *exec_path, char **exec_args)
{
	#ifndef NOFORK
	int fd[2];
	struct stat st;

	if (stat(exec_path, &st) == -1) {
		anubis_error("stat() failed. %s", strerror(errno));
		return -1;
	}
	info(VERBOSE, "Executing a local program (%s)...", exec_path);
	if (make_sockets(fd))
		return -1;

	signal(SIGCHLD, sig_local);
	switch(fork())
	{
		case -1: /* an error */
			anubis_error("fork() failed.");
			close_socket(fd[0]);
			close_socket(fd[1]);
			return -1;
		case  0: /* a child process */
			close_socket(fd[0]);
			dup2(fd[1], 0);
			dup2(fd[1], 1);
			close_socket(fd[1]);
			execvp(exec_path, exec_args);
			anubis_error("execvp() failed."); /* execvp() failed */
			return -1;
	}
	close_socket(fd[1]);
	return fd[0];
	#else
	 info(NORMAL, "NOFORK is set. Couldn't execute a local program.");
	 return -1;
	#endif /* not NOFORK */
}

/*******************
  Bind and listen.
********************/

int bind_and_listen(char *host, unsigned int port)
{
	int sd = 0;
	struct sockaddr_in addr;

	if ((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		anubis_error("Can't create stream socket.");

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	if (topt & T_NAMES) {
		struct hostent *hp = 0;
		if ((hp = gethostbyname(host)) == 0) {
			addr.sin_addr.s_addr = inet_addr(host);
			if ((int)addr.sin_addr.s_addr == -1)
				anubis_error("Unknown host %s", host);
		}
		else
			memcpy((char *)&addr.sin_addr.s_addr, hp->h_addr, hp->h_length);
	}
	else
		addr.sin_addr.s_addr = htonl(INADDR_ANY);

	addr.sin_port = htons(port);

	if (bind(sd, (struct sockaddr *)&addr, sizeof(addr)))
		anubis_error("bind() failed. %s", strerror(errno));
	info(VERBOSE, "Anubis bound to %s:%u", inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
	if (listen(sd, 5))
		anubis_error("listen() failed. %s", strerror(errno));

	return sd;
}

/**************************
   Send a data to server
***************************/

int swrite_char(int method, void *sd, char ch)
{
	if (method == CLIENT) {
		#ifndef NOSSL
		if (popt & P_SSL_CLIENT) {
			if (SSL_write((SSL *)sd, &ch, 1) <= 0)
				return -1;
		} else
		#endif /* not NOSSL */
		if (!(popt & P_SSL_CLIENT)) {
			if (send((int)sd, &ch, 1, 0) == -1)
				return -1;
		}
	}
	else if (method == SERVER) {
		#ifndef NOSSL
		if (popt & P_SSL_SERVER) {
			if (SSL_write((SSL *)sd, &ch, 1) <= 0)
				return -1;
		} else
		#endif /* not NOSSL */
		if ((int)sd == -1) { /* stdout */
			write(1, &ch, 1);
		}
		else if (!(popt & P_SSL_SERVER)) {
			if (send((int)sd, &ch, 1, 0) == -1)
				return -1;
		}
	}
	bytes_sent++;
	return 0;
}

void swrite(int method, void *sd, char *txt)
{
	int len = strlen(txt);

	if (method == CLIENT) {
		#ifndef NOSSL
		if (popt & P_SSL_CLIENT) {
			if (SSL_write((SSL *)sd, txt, len) <= 0)
				socket_error();
		} else
		#endif /* not NOSSL */
		if (!(popt & P_SSL_CLIENT)) {
			if (send((int)sd, txt, len, 0) == -1)
				socket_error();
		}
	}
	else if (method == SERVER) {
		#ifndef NOSSL
		if (popt & P_SSL_SERVER) {
			if (SSL_write((SSL *)sd, txt, len) <= 0)
				socket_error();
		} else
		#endif /* not NOSSL */
		if ((int)sd == -1) { /* stdout */
			write(1, txt, len);
		}
		else if (!(popt & P_SSL_SERVER)) {
			if (send((int)sd, txt, len, 0) == -1)
				socket_error();
		}
	}
	bytes_sent += len;
	return;
}

void sendline(void *sd, char *txt) /* control CRLF state; be sure to send CRLF */
{
	remcrlf(txt);
	swrite(CLIENT, sd, txt);
	swrite(CLIENT, sd, CRLF);
	return;
}

/**************************
  Read a data from server
***************************/

static int mread(int method, void *sd, char *ptr)
{
	static int nread = 0;
	static char *read_ptr = 0;
	static char buf[LINEBUFFER];

	if (nread <= 0) {
	again:
		if (method == CLIENT) {
			#ifndef NOSSL
			if (popt & P_SSL_CLIENT) {
				nread = SSL_read((SSL *)sd, buf, LINEBUFFER);
			} else
			#endif /* not NOSSL */
			if (!(popt & P_SSL_CLIENT)) {
				nread = recv((int)sd, buf, LINEBUFFER, 0);
			}
		}
		else if (method == SERVER) {
			#ifndef NOSSL
			if (popt & P_SSL_SERVER) {
				nread = SSL_read((SSL *)sd, buf, LINEBUFFER);
			} else
			#endif /* not NOSSL */
			if ((int)sd == -1) { /* stdin */
				nread = read(0, buf, LINEBUFFER);
			}
			else if (!(popt & P_SSL_SERVER)) {
				nread = recv((int)sd, buf, LINEBUFFER, 0);
			}
		}
		if (nread < 0) {
			if (errno == EINTR)
				goto again;
			return -1;
		}
		else if (nread == 0)
			return 0;
		read_ptr = buf;
	}
	nread--;
	*ptr = *read_ptr++;
	return 1;
}

int recvline(int method, void *sd, void *vptr, int maxlen)
{
	int n, rc;
	char c, *ptr;

	ptr = vptr;
	for (n = 1; n < maxlen; n++)
	{
		if ((rc = mread(method, sd, &c)) == 1) {
			*ptr++ = c;
			if (c == '\n')
				break;
		}
		else if (rc == 0) {
			if (n == 1)
				return 0;
			else
				break;
		}
		else
			return -1;
	}
	*ptr = 0;
	return n;
}

/*******************
   Get a response
********************/

void get_response_smtp(int method, void *sd, char *buf, int size)
{
	int n;
	char line[LINEBUFFER];

	if (buf != 0)
		memset(buf, 0, size);
	do {
		n = recvline(method, sd, line, LINEBUFFER);
		if (buf != 0) {
			strncat(buf, line, size);
			size -= n;
		}
	} while (line[3] == '-');
	#ifdef DEBUG_CODE
	if (output == DEBUG)
		debug_response(buf);
	#endif /* DEBUG_CODE */
	return;
}

/******************************
   Close a socket descriptor
*******************************/

void close_socket(int sd)
{
	if (sd)
		close(sd);
	return;
}

/* EOF */

