/*
    pgp.c

    This file is a part of Anubis.
    Copyright (C) 2001, 2002 The Anubis Team.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "headers.h"
#include "extern.h"

#ifndef NOGPG
#include <gpgme.h>

static void init_gpgme(void);
static char *pgp_encrypt_to_remailer(char *);
static char *pgp_encrypt_to_users(char *);
static char *pgp_sign(char *);

#define fail_if_err(a)	do {									\
					if (a) {							\
						anubis_error("GpgmeError: %s",	\
						gpgme_strerror(a));				\
					}									\
				} while(0)

static void init_gpgme(void)
{
	if (gpgme_check_engine())
		anubis_error("GPGME: Engine init failed.");
	return;
}

static const char *passphrase_cb(void *opaque, const char *desc, void *r_hd)
{
	static char pgp_password[1024];
	char *my_desc = 0;
	int begin = 0;
	int end = 0;
	int i = 0;
	int j = 0;

	if (!desc)
		return 0;

	memset(pgp_password, 0, sizeof(pgp_password));
	if (pgp_sign_password) {
		if (strlen(pgp_sign_password) != 0) {
			safe_strcpy(pgp_password, pgp_sign_password);
			return pgp_password;
		}
	}
	my_desc = (char *)xmalloc(strlen(desc) + 1);

	for (i = 0; i <= strlen(desc); i++)
	{
		if (desc[i] == ' ') {
			begin = i + 1;
			break;
		}
	}

	for (i = 0; i <= strlen(desc); i++)
	{
		if (desc[i] == '\n')
			end = i - 1;
	}

	for (i = begin; i <= end; i++)
		my_desc[j++] = desc[i];

	my_desc[j] = '\0';
	info(NORMAL, "Requesting passphrase for %s", my_desc);
	free(my_desc);

	auth_password("Passphrase: ", pgp_password, sizeof(pgp_password));
	if (strlen(pgp_password) == 0) /* an empty passphrase */
		strcpy(pgp_password, " \n");
	return pgp_password;
}

static char *pgp_encrypt_to_users(char *pgp_data)
{
	GpgmeCtx ctx;
	GpgmeError err;
	GpgmeData in, out;
	GpgmeRecipients rset;
	char buf[100];
	char *encrypted_data;
	int pgp_keys_i, pgp_keys_j = 0;
	char pgp_key[100];
	size_t nread;

	init_gpgme();

	encrypted_data = (char *)xmalloc(strlen(pgp_data) + 1024);
	memset(buf, 0, sizeof(buf));
	memset(pgp_key, 0, sizeof(pgp_key));

	fail_if_err(gpgme_new(&ctx));
	gpgme_set_armor(ctx, 1);
	fail_if_err(gpgme_data_new_from_mem(&in, pgp_data, strlen(pgp_data), 0));
	fail_if_err(gpgme_data_new(&out));
	fail_if_err(gpgme_recipients_new(&rset));

	strncat(pgp_keys, ",", 1);
	for (pgp_keys_i = 0; pgp_keys_i <= strlen(pgp_keys); pgp_keys_i++)
	{
		if (pgp_keys[pgp_keys_i] == ',') { /* comma found, so add KEY-ID */
			fail_if_err(gpgme_recipients_add_name_with_validity(rset, pgp_key,
			GPGME_VALIDITY_FULL));
			memset(pgp_key, 0, sizeof(pgp_key));
			pgp_keys_j = 0;
		}
		else /* it is not a comma, so add char to KEY_ID string */
			pgp_key[pgp_keys_j++] = pgp_keys[pgp_keys_i];
	}
	fail_if_err(gpgme_op_encrypt(ctx, rset, in, out));
	fflush(0);

	fail_if_err(gpgme_data_rewind(out));

	while (!(err = gpgme_data_read(out, buf, 100, &nread)))
	{
		strncat(encrypted_data, buf, nread);
		memset(buf, 0, sizeof(buf));
	}
	if (err != GPGME_EOF)
		fail_if_err(err);

	gpgme_data_release(out);
	gpgme_data_rewind(in);
	gpgme_data_release(in);
	gpgme_release(ctx);

	return (char *)encrypted_data;
}

static char *pgp_sign(char *pgp_data)
{
	GpgmeCtx ctx;
	GpgmeError err;
	GpgmeData in, out;
	char buf[100];
	char *signed_data;
	size_t nread;

	init_gpgme();

	signed_data = (char *)xmalloc(strlen(pgp_data) + 1024);
	memset(buf, 0, sizeof(buf));

	fail_if_err(gpgme_new(&ctx));
	if (!(getenv("GPG_AGENT_INFO")))
		gpgme_set_passphrase_cb(ctx, passphrase_cb, 0);

	gpgme_set_textmode(ctx, 1);
	gpgme_set_armor(ctx, 1);

	fail_if_err(gpgme_data_new_from_mem(&in, pgp_data, strlen(pgp_data), 0));
	fail_if_err(gpgme_data_new(&out));
	fail_if_err(gpgme_op_sign(ctx, in, out, GPGME_SIG_MODE_CLEAR));
	fflush(0);

	fail_if_err(gpgme_data_rewind(out));
	while (!(err = gpgme_data_read(out, buf, 100, &nread)))
	{
		strncat(signed_data, buf, nread);
		memset(buf, 0, sizeof(buf));
	}
	if (err != GPGME_EOF)
		fail_if_err(err);

	gpgme_data_release(out);
	gpgme_data_rewind(in);
	gpgme_data_release(in);
	gpgme_release(ctx);

	return (char *)signed_data;
}

static char *pgp_encrypt_to_remailer(char *pgp_data)
{
	#define EXTRA_PGP_BUF 8192
	GpgmeCtx ctx;
	GpgmeError err;
	GpgmeData in, out;
	GpgmeRecipients rset;
	char buf[100];
	char *data_to_encrypt;
	size_t nread;
	char mail_buf[1024];

	init_gpgme();

	data_to_encrypt = (char *)xmalloc(strlen(pgp_data) + EXTRA_PGP_BUF);
	memset(buf, 0, sizeof(buf));

	fail_if_err(gpgme_new(&ctx));
	gpgme_set_armor(ctx, 1);

	/*
	   Remailer Type-I support (with PGP encryption).
	*/

	if (mopt & M_RMRRT)
		sprintf(data_to_encrypt, "\r\n::\r\nAnon-To: %s\r\n", rm_rrt);
	else if (mopt & M_RMPOST)
		sprintf(data_to_encrypt, "\r\n::\r\nAnon-Post-To: %s\r\n", rm_post);
	if ((mopt & M_RMLT) || (mopt & M_RMRLT)) {
		sprintf(mail_buf, "Latent-Time: +%s%s\r\n",
		rm_lt, (mopt & M_RMRLT) ? "r" : "");
		strncat(data_to_encrypt, mail_buf, strlen(mail_buf));
	}
	if (!(topt & T_TUNNEL)) {
		strncat(data_to_encrypt, "\r\n##\r\n", 6);
		if (mopt & M_RMHEADER) {
			sprintf(mail_buf, "%s\r\n", rm_header);
			strncat(data_to_encrypt, mail_buf, strlen(mail_buf));
		}
		sprintf(mail_buf, "Subject: %s\r\n" , subject);
		strncat(data_to_encrypt, mail_buf, strlen(mail_buf));
		strncat(data_to_encrypt, "MIME-Version: 1.0\r\n", 20);
		sprintf(mail_buf, "Content-Type: text/plain; charset=%s\r\n"
						"Content-Transfer-Encoding: %s\r\n\r\n",
						(mopt & M_CHARSET) ? charset : "US-ASCII",
						(mopt & M_CHARSET) ? "8bit" : "7bit");
		strncat(data_to_encrypt, mail_buf, strlen(mail_buf));
	}
	strncat(data_to_encrypt, pgp_data, strlen(pgp_data));
	memset(buf, 0, sizeof(buf));

	fail_if_err(gpgme_data_new_from_mem(&in, data_to_encrypt, strlen(data_to_encrypt), 0));
	fail_if_err(gpgme_data_new(&out));
	fail_if_err(gpgme_recipients_new(&rset));
	fail_if_err(gpgme_recipients_add_name_with_validity(rset, rm_pgp_key,
	GPGME_VALIDITY_FULL));
	fail_if_err(gpgme_op_encrypt(ctx, rset, in, out));
	fflush(0);

	fail_if_err(gpgme_data_rewind(out));
	memset(data_to_encrypt, 0, strlen(pgp_data) + EXTRA_PGP_BUF);

	while (!(err = gpgme_data_read(out, buf, 100, &nread)))
	{
		strncat(data_to_encrypt, buf, nread);
		memset(buf, 0, sizeof(buf));
	}
	if (err != GPGME_EOF)
		fail_if_err(err);

	gpgme_recipients_release(rset);
	gpgme_data_release(in);
	gpgme_data_release(out);
	gpgme_release(ctx);

	return (char *)data_to_encrypt;
}

void check_pgp(void)
{
	unsigned long size = 0;
	char *homedir = get_homedir(local_user);

	setenv("HOME", get_homedir(remote_user), 1);
	strncat(msgbody, CRLF, 2);

	if (mopt & M_PGPSIGN) { /* Sign a message */
		char *tmp = pgp_sign(msgbody);
		if (popt & P_ERROR)
			return;
		size = strlen(tmp) + 1;
		msgbody = (char *)xrealloc((char *)msgbody, size);
		memset(msgbody, 0, size);
		strncpy(msgbody, tmp, size - 1);
		if (tmp)
			free(tmp);
	}
	if (mopt & M_PGP) { /* Encrypt a message body */
		char *tmp = pgp_encrypt_to_users(msgbody);
		if (popt & P_ERROR)
			return;
		size = strlen(tmp) + 1;
		msgbody = (char *)xrealloc((char *)msgbody, size);
		memset(msgbody, 0, size);
		strncpy(msgbody, tmp, size - 1);
		if (tmp)
			free(tmp);
	}
	if ((mopt & M_RMPGP) && !(mopt & M_RMPGPFILE)) { /* Remailer Type-I */
		char *tmp = pgp_encrypt_to_remailer(msgbody);
		if (popt & P_ERROR)
			return;
		size = strlen(tmp) + 1;
		msgbody = (char *)xrealloc((char *)msgbody, size);
		memset(msgbody, 0, size);
		strncpy(msgbody, tmp, size - 1);
		if (tmp)
			free(tmp);
	}
	setenv("HOME", homedir, 1);
	if (pgp_sign_password) {
		xfree(pgp_sign_password);
	}
	return;
}

#endif /* NOGPG */

/* EOF */

