/*
    rc.c

    This file is a part of Anubis.
    Copyright (C) 2001, 2002 The Anubis Team.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "headers.h"
#include "extern.h"

static void match_options_common(char *);
#if !defined(NOREGEX) || defined(PCRE)
 static int  get_regex(int, char *, char *, int);
 static char *parse_line_regex(char *);
 static void get_action_line(char *);
 static void match_action(char *, char *);
#endif /* not NOREGEX || PCRE */

#define set(a, b, c) \
	if (strncmp((char *)a, "yes", 3) == 0) { \
		b |= c; \
		return; \
	} \
	else if (strncmp((char *)a, "no", 2) == 0) { \
		b &= ~c; \
		return; \
	} \
	else \
		return;

#define if_empty_quit(a) \
	if (strlen((char *)a) == 0) \
		return;

#define if_empty_set(a, b, c) \
	if (strlen((char *)a) == 0) { \
		b &= ~c; \
		return; \
	} \
	else if (strlen((char *)a) != 0) { \
		b |= c; \
		return; \
	}

void open_config_file(int method, char *user)
{
	char *config_path = 0;

	if (method == LOCAL && getuid() == 0 && getgid() == 0) {
		if (popt & P_ALTRC)
			config_path = altrc;
		else
			config_path = allocbuf(DEFAULT_GLOBAL_CONFIG_FILE, MAXPATHLEN);
	}
	else {
		int n;
		char *home = 0;
		char *config_file = (char *)DEFAULT_LOCAL_CONFIG_FILE;

		if (popt & P_ALTRC)
			config_path = altrc;
		else {
			home = get_homedir(user);
			if (home == 0)
				home = (char *)"";

			n = strlen(home) + strlen(config_file) + 2;
			n = n > MAXPATHLEN ? MAXPATHLEN : n + 1;
			config_path = (char *)xmalloc(n);
			#ifdef HAVE_SNPRINTF
			snprintf(config_path, n - 1,
			#else
			sprintf(config_path,
			#endif /* HAVE_SNPRINTF */
			"%s/%s", home, config_file);
		}
	}
	if (check_file_mode(config_path) == -1) { /* If wrong permissions... */
		if (!(popt & P_ALTRC)) {
			xfree(config_path);
		}
		return;
	}
	#ifndef NOFORK
	if (method == LOCAL) {
		struct stat st;
		if (stat(config_path, &st) == 0) {
			if (global_mtime != st.st_mtime)
				global_mtime = st.st_mtime;
			else {
				if (!(popt & P_ALTRC)) {
					xfree(config_path);
				}
				return;
			}
		}
	}
	#endif /* not NOFORK */
	if ((fp_config = fopen(config_path, "r")) == 0) {
		if (popt & P_ALTRC) {
			anubis_error("Anubis RC file error: %s", strerror(errno));
			popt &= ~P_ERROR;
		}
	}
	if (!(popt & P_ALTRC))
		xfree(config_path);
	return;
}

void read_config_file(void)
{
	char rcline[LINEBUFFER];

	if (fp_config == 0)
		return;

	if (topt & T_TUNNEL) /* The Tunnel */
		config_position = get_position(BEGIN_TUNNEL);
	#ifndef TUNNEL_OPTIMIZE
	else if (popt & P_SMTP) /* The SMTP client */
		config_position = get_position(BEGIN_SMTP);
	#endif /* not TUNNEL_OPTIMIZE */

	fseek(fp_config, config_position, SEEK_SET);
	while (fgets(rcline, LINEBUFFER, fp_config) != 0)
	{
		if (strncmp(rcline, "#", 1) == 0 || strncmp(rcline, "\n", 1) == 0)
			continue; /* skip a comment or an empty line */
		else if (strncmp(rcline, END_CONFIG, strlen(END_CONFIG)) == 0)
			break; /* THE END */
		else
			match_options_common(rcline);
	}
	return;
}

char *parse_line_option(char *line)
{
	char *ptr = 0;
	ptr = strchr(line, '=');

	if (ptr == 0)
		return line;
	do {
		ptr++;
	} while (*ptr == ' ' || *ptr == '\t');

	remcrlf(ptr);
	return ptr;
}

static void match_options_common(char *rcline)
{
	char *ptr = 0;
	char buf[LINEBUFFER];

	safe_strcpy(buf, rcline);
	change_to_lower(buf);

	/*
	   Generic options.
	*/

	if (strncmp(buf, "output", 6) == 0) {
		ptr = parse_line_option(buf);
		if (strncmp(ptr, "silent", 6) == 0)
			output = SILENT;
		else if (strncmp(ptr, "normal", 6) == 0)
			output = NORMAL;
		else if (strncmp(ptr, "verbose", 7) == 0)
			output = VERBOSE;
		#ifdef DEBUG_CODE
		else if (strncmp(ptr, "debug", 5) == 0)
			output = DEBUG;
		#endif /* DEBUG_CODE */
		else {
			anubis_error("Anubis RC file error: unknown output mode.");
			popt &= ~P_ERROR;
			output = NORMAL;
		}
		return;
	}

	if (strncmp(buf, "server", 6) == 0 || strncmp(buf, "host", 4) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		parse_hostname_port(ptr, hostname, &portnumber);
		return;
	}

	/*
	   SOCKS Proxy.
	*/

	if (strncmp(buf, "socks-proxy", 11) == 0) {
		parse_hostname_port(parse_line_option(rcline), socks_proxy, &socks_port);
		if_empty_set(socks_proxy, popt, P_SOCKS);
	}
	if (strncmp(buf, "socks-v4", 8) == 0) {
		set(parse_line_option(buf), popt, P_SOCKSV4);
	}
	if (strncmp(buf, "socks-username", 14) == 0) {
		safe_strcpy(socks_username, parse_line_option(rcline));
		if_empty_set(socks_username, popt, P_SOCKSUSERNAME);
	}
	if (strncmp(buf, "socks-password", 14) == 0) {
		safe_strcpy(socks_password, parse_line_option(rcline));
		if_empty_set(socks_password, popt, P_SOCKSPASSWORD);
	}

	/*
	   WinGates.
	*/

	if (strncmp(buf, "wingate1", 8) == 0) {
		parse_hostname_port(parse_line_option(rcline), wingate1, &wingate1_port);
		if_empty_set(wingate1, popt, P_WG1);
	}
	if (strncmp(buf, "wingate2", 8) == 0) {
		parse_hostname_port(parse_line_option(rcline), wingate2, &wingate2_port);
		if_empty_set(wingate2, popt, P_WG2);
	}

	/*
	   The SMTP session.
	*/

	#ifndef TUNNEL_OPTIMIZE
	if (popt & P_SMTP) {
		if (strncmp(buf, "realuser", 8) == 0) {
			set(parse_line_option(buf), popt, P_REALUSER);
		}
		if (strncmp(buf, "auth-username", 13) == 0) {
			safe_strcpy(username, parse_line_option(rcline));
			if_empty_set(username, popt, P_USERNAME);
		}
		if (strncmp(buf, "auth-password", 13) == 0) {
			safe_strcpy(password, parse_line_option(rcline));
			memset(rcline, 0, strlen(password));
			if_empty_quit(password);
			popt |= P_PASSWORD;
			popt |= P_SMTPAUTH;
			return;
		}
	}
	#endif /* not TUNNEL_OPTIMIZE */

	/*
	   The SMTP Tunnel.
	*/

	if (topt & T_TUNNEL) {
		if (strncmp(buf, "bind", 4) == 0) {
			ptr = parse_line_option(rcline);
			if_empty_quit(ptr);
			parse_hostname_host(ptr, tunnel_hostname, &tunnel_portnumber);
			popt &= ~P_SMTP;
			topt |= T_TUNNEL;
			if (strlen(tunnel_hostname) != 0)
				topt |= T_NAMES;
			return;
		}
		if (strncmp(buf, "local-mta", 9) == 0) {
			#ifndef NOFORK
			ptr = parse_line_option(rcline);
			if_empty_quit(ptr);
			xfree(execpath);
			execpath = allocbuf(ptr, MAXPATHLEN);
			topt |= T_LOCAL_MTA;
			#else
			info(NORMAL, "Local mode not supported under this OS.");
			#endif /* not NOFORK */
			return;
		}
		#ifndef NOFORK
		if (strncmp(buf, "mta-args", 8) == 0) {
			ptr = parse_line_option(rcline);
			if_empty_quit(ptr);
			if (topt & T_RCEXECARGS)
				free_execargs(execargs);
			execargs = gen_execargs(ptr);
			topt |= T_RCEXECARGS;
			return;
		}
		#endif /* not NOFORK */
		if (strncmp(buf, "auth-local", 10) == 0) {
			set(parse_line_option(buf), topt, T_AUTH_LOCAL);
		}
		if (strncmp(buf, "ident-only", 10) == 0) {
			set(parse_line_option(buf), topt, T_IDENT_ONLY);
		}
		#ifndef NOFORK
		if (strncmp(buf, "setremoteuser", 13) == 0) {
			set(parse_line_option(buf), topt, T_SETREMOTE_USER);
		}
		#endif /* not NOFORK */
	}

	/*
	   TLS/SSL.
	*/

	#ifndef NOSSL
	if (strncmp(buf, "ssl", 3) == 0) {
		ptr = parse_line_option(buf);
		if (strncmp(ptr, "yes", 3) == 0)
			popt &= ~P_SSL_NO;
		else if (strncmp(ptr, "no", 2) == 0)
			popt |= P_SSL_NO;
		else if (strncmp(ptr, "only", 4) == 0)
			popt |= P_SSL_ONLY;
		return;
	}
	if (topt & T_TUNNEL) {
		if (strncmp(buf, "cert", 4) == 0) {
			ptr = parse_line_option(rcline);
			if_empty_quit(ptr);
			xfree(ssl_cert);
			ssl_cert = allocbuf(ptr, MAXPATHLEN);
			return;
		}
		if (strncmp(buf, "key", 3) == 0) {
			ptr = parse_line_option(rcline);
			if_empty_quit(ptr);
			xfree(ssl_key);
			ssl_key = allocbuf(ptr, MAXPATHLEN);
			return;
		}
	}
	#endif /* not NOSSL */

	return;
}

/******************************
  Regular Expressions support
*******************************/

#if !defined(NOREGEX) || defined(PCRE)

int read_regex_block(int method, char *regex, int size)
{
	char rcline[LINEBUFFER];

	if (fp_config == 0)
		return 0;

	while (fgets(rcline, LINEBUFFER, fp_config) != 0)
	{
		if (strncmp(rcline, "#", 1) == 0 || strncmp(rcline, "\n", 1) == 0)
			continue; /* skip a comment or an empty line */
		else if (strncmp(rcline, END_CONFIG, strlen(END_CONFIG)) == 0)
			break; /* END OF REGEX BLOCK */
		else {
			if (get_regex(method, rcline, regex, size) == 0)
				continue;
			else
				return 1;
		}
	}
	return 0;
}

static int get_regex(int method, char *rcline, char *regex, int size)
{
	char *ptr = 0;

	if (method == HEADER) { /* "IF HEADER =..." */
		if (regex_match("[ \t]*[Ii][Ff][ \t]+[Hh][Ee][Aa][Dd][Ee][Rr][ \t]+=", rcline)) {
			ptr = parse_line_regex(rcline);
			if (strlen(ptr) == 0)
				return 0;
			strncpy(regex, ptr, size);
			return 1;
		}
	}
	else if (method == COMMAND) { /* "IF COMMAND =..." */
		if (regex_match("[ \t]*[Ii][Ff][ \t]+[Cc][Oo][Mm][Mm][Aa][Nn][Dd][ \t]+=", rcline)) {
			ptr = parse_line_regex(rcline);
			if (strlen(ptr) == 0)
				return 0;
			strncpy(regex, ptr, size);
			return 1;
		}
	}
	return 0;
}

static char *parse_line_regex(char *rcline)
{
	char *ptr = 0;
	ptr = strchr(rcline, '=');
	if (ptr == 0)
		return rcline;
	ptr++;
	remcrlf(ptr);
	return ptr;
}

/**************************
   Read the ACTION-BLOCK
***************************/

int read_action_block(char *line)
{
	char rcline[LINEBUFFER];

	if (fp_config == 0)
		return 0;

	while (fgets(rcline, LINEBUFFER, fp_config) != 0)
	{
		if (strncmp(rcline, "#", 1) == 0 || strncmp(rcline, "\n", 1) == 0
		|| strncmp(rcline, "{", 1) == 0)
			continue; /* skip a comment, an empty line, and "{" */
		else if (regex_match("[ \t]*}[ \t]*", rcline)) /* a single '}' character */
			break; /* END OF REGEX BLOCK ACTION */
		else {
			get_action_line(rcline);
			match_action(rcline, line);
			return 1;
		}
	}
	return 0;
}

static void get_action_line(char *rcline)
{
	char *ptr = 0;
	char buf[LINEBUFFER];

	safe_strcpy(buf, rcline);
	ptr = buf;
	while (*ptr == ' ' || *ptr == '\t')
		ptr++;

	remcrlf(ptr);
	strncpy(rcline, ptr, LINEBUFFER);
	return;
}

static void match_action(char *rcline, char *line)
{
	char *ptr = 0;
	char buf[LINEBUFFER];

	safe_strcpy(buf, rcline);
	change_to_lower(buf);

	if (strncmp(buf, "add", 3) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		addlist = (char *)xrealloc((char *)addlist,
		strlen(addlist) + strlen(ptr) + 4);
		strncat(addlist, ptr, strlen(ptr));
		strncat(addlist, CRLF, 2);
		return;
	}
	if (strncmp(buf, "remove", 6) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		remlist = (char *)xrealloc((char *)remlist,
		strlen(remlist) + strlen(ptr) + 4);
		strncat(remlist, ptr, strlen(ptr));
		strncat(remlist, CRLF, 2);
		return;
	}
	if (strncmp(buf, "modify", 6) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		strncpy(line, ptr, LINEBUFFER);
		strncat(line, CRLF, 2);
		return;
	}

	if (strncmp(buf, "rot13all", 5) == 0) {
		set(parse_line_option(buf), mopt, M_ROT13);
	}
	if (strncmp(buf, "rot13subject", 12) == 0) {
		set(parse_line_option(buf), mopt, M_ROT13S);
	}
	if (strncmp(buf, "rot13body", 9) == 0) {
		set(parse_line_option(buf), mopt, M_ROT13B);
	}
	if (strncmp(buf, "signature", 9) == 0) {
		set(parse_line_option(buf), mopt, M_SIGNATURE);
	}
	if (strncmp(buf, "body-file", 9) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(body_file);
		body_file = allocbuf(ptr, MAXPATHLEN);
		xfree(msgbody);
		msgbody = (char *)xmalloc(1);
		mopt |= M_BODYATTACH;
		mopt |= M_BODY;
		return;
	}

	/*
	   GnuPG support.
	*/

	#ifndef NOGPG
	if (strncmp(buf, "pgp-encrypt", 11) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(pgp_keys);
		pgp_keys = allocbuf(ptr, 0);
		mopt |= M_PGP;
		return;
	}
	if (strncmp(buf, "pgp-sign", 8) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(pgp_sign_password);
		pgp_sign_password = allocbuf(ptr, 0);
		mopt |= M_PGPSIGN;
		return;
	}
	#endif /* not NOGPG */

	/*
	   Remailer Type-I support.
	*/

	if (strncmp(buf, "rm-rrt", 6) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(rm_rrt);
		rm_rrt = allocbuf(ptr, 0);
		mopt |= M_RM;
		mopt |= M_RMRRT;
		return;
	}
	if (strncmp(buf, "rm-post", 7) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(rm_post);
		rm_post = allocbuf(ptr, 0);
		mopt |= M_RM;
		mopt |= M_RMPOST;
		return;
	}
	#ifndef NOGPG
	if (strncmp(buf, "rm-pgp", 6) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(rm_pgp_key);
		rm_pgp_key = allocbuf(ptr, 0);
		mopt |= M_RMPGP;
		return;
	}
	#endif /* not NOGPG */
	if (strncmp(buf, "rm-header", 9) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		xfree(rm_header);
		rm_header = allocbuf(ptr, 512);
		mopt |= M_RMHEADER;
		return;
	}
	if (strncmp(buf, "rm-lt", 5) == 0) {
		ptr = parse_line_option(rcline);
		if_empty_quit(ptr);
		safe_strcpy(rm_lt, ptr);
		mopt |= M_RMLT;
		return;
	}
	if (strncmp(buf, "rm-rlt", 6) == 0) {
		set(parse_line_option(buf), mopt, M_RMRLT);
	}
	return;
}

#endif /* not NOREGEX || PCRE */

/****************************
  Find the 'LINE' position.
*****************************/

unsigned long get_position(char *line)
{
	char buf[LINEBUFFER];

	if (fp_config == 0)
		return 0;

	rewind(fp_config);
	while (fgets(buf, LINEBUFFER, fp_config) != 0)
	{
		if (strncmp(buf, line, strlen(line)) == 0)
			return (unsigned long)ftell(fp_config);
	}
	return 0;
}

/* EOF */

