/*
    smtp.c

    This file is a part of Anubis.
    Copyright (C) 2001, 2002 The Anubis Team.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "headers.h"
#include "extern.h"

#ifndef TUNNEL_OPTIMIZE

static void smtp_client_start(void *);
static void say_smtp_hello(void *);
static void say_smtp(void *, char *, char *, ...);
static void check_smtp_response(void *, char *);
static void esmtp_auth(void *, char *);
static void build_message(void *);
static void attach_file(int, void *);
static void remailer_cypherpunk(int, void *);
static void remailer_mixmaster(void);
static void smtp_failed(void *);
static void smtp_success(void);

void smtp_client_init(void)
{
	int sockfd;
	info(VERBOSE, "Initializing the SMTP session...");

	check_all_files(local_user);
	check_rot13();
	#ifndef NOGPG
	check_pgp();
	#endif /* not NOGPG */

	/*
           The Mixmaster support. Execute a local Mixmaster program.
	*/

	if (mopt & M_MIXMASTER) {
		remailer_mixmaster();
		smtp_success();
		quit(0);
	}
	sockfd = make_remote_connection(hostname, portnumber);
	remote_client = (void *)sockfd;

	smtp_client_start(remote_client);
	close_socket(sockfd);
	smtp_success();
	quit(0);
}

static void smtp_client_start(void *sd)
{
	/*
	   Get a welcome message.
	*/

	get_response_smtp(CLIENT, sd, 0, 0);

	/*
	   Transmit a data.
	*/

	say_smtp_hello(sd);
	sd = remote_client;
	info(NORMAL, "Sending a message...");

	if (mopt & M_NOROOT)
		say_smtp(sd, "MAIL", "%s from:<%s>", smtp_mode, parse_address(from));
	else
		say_smtp(sd, "MAIL", "%s from:<root@%s>", smtp_mode, hostname);

	say_smtp(sd, "RCPT", "rcpt to:<%s>", parse_address(to));
	if (mopt & M_CC) { /* Carbon Copy */
		char *ptr = strtok(carbon_copy, ",");
		say_smtp(sd, "RCPT", "rcpt to:<%s>", parse_address(ptr));
		for (;;)
		{
			ptr = strtok(0, ",");
			if (ptr == 0)
				break;
			say_smtp(sd, "RCPT", "rcpt to:<%s>", parse_address(ptr));
		}
	}
	if (mopt & M_BCC) { /* Blind Carbon Copy */
		char *ptr = strtok(blind_carbon_copy, ",");
		say_smtp(sd, "RCPT", "rcpt to:<%s>", parse_address(ptr));
		for (;;)
		{
			ptr = strtok(0, ",");
			if (ptr == 0)
				break;
			say_smtp(sd, "RCPT", "rcpt to:<%s>", parse_address(ptr));
		}
	}
	say_smtp(sd, "DATA", "data");
	build_message(sd);
	say_smtp(sd, "QUIT", "quit");

	#ifndef NOSSL
	end_ssl(CLIENT, ssl_client, ctx_client);
	ssl_client = 0;
	#endif /* not NOSSL */

	return;
}

static void say_smtp_hello(void *sd)
{
	if (popt & P_OVERLOAD) { /* Old Sendmail HELO bug */
		int c;
		swrite(CLIENT, sd, "helo ");
		for (c = 0; c < 50; c++) /* 1200B > 1024B :) */
			swrite(CLIENT, sd, "xxxxxxxxxxxxxxxxxxxxxxxx");
		say_smtp(sd, "HELO", CRLF);
	}
	else
		say_smtp(sd, "EHLO", "ehlo %s", hostname);
	return;
}

static void say_smtp(void *sd, char *name, char *format, ...)
{
	va_list arglist;
	char buf[1024];
	memset(buf, 0, sizeof(buf));

	va_start(arglist, format);
	#ifdef HAVE_VSNPRINTF
	vsnprintf(buf, 1024, format, arglist);
	#else
	vsprintf(buf, format, arglist);
	#endif /* HAVE_VSNPRINTF */
	va_end(arglist);

	sendline(sd, buf);
	check_smtp_response(sd, name);
	return;
}

static void parse_ehlo_reply(void *sd, char *buf)
{
	/*
	   ESMTP TLS/SSL support.
	*/

	#ifndef NOSSL
	if (strstr(buf, "STARTTLS") && !(popt & P_SSL_CLIENT) && !(popt & P_SSL_NO)) {
		say_smtp(sd, "STLS", "starttls"); /* The STARTTLS command */
		ctx_client = init_ssl_client();
		ssl_client = start_ssl_client((int)sd, ctx_client);
		sd = remote_client = (void *)ssl_client;
		say_smtp(sd, "EHLO", "ehlo %s", hostname); /* after TLS/SSL negotiation */
		return;
	}
	else if (!(strstr(buf, "STARTTLS")) && !(popt & P_SSL_CLIENT) && !(popt & P_SSL_NO)) {
		info(NORMAL, "ESMTP TLS/SSL is unavailable at %s:%u", hostname, portnumber);
		if (popt & P_SSL_ONLY)
			quit(0);
		else
			info(NORMAL, "Continuing...");
	}
	#endif /* not NOSSL */

	/*
	   ESMTP Authentication.
	*/

	if ((popt & P_SMTPAUTH) && (strstr(buf, "AUTH "))) {
		if (!(popt & P_USERNAME))
			auth_username(username, sizeof(username));
		if (strncmp(username, "root\0", 5) == 0)
			anubis_error("Cannot accept 'root' as the user name.");
		if (!(popt & P_PASSWORD))
			auth_password("Password: ", password, sizeof(password));
		esmtp_auth(sd, buf);
	}
	else if ((popt & P_SMTPAUTH) && !(strstr(buf, "AUTH ")))
		info(NORMAL, "ESMTP AUTH is unavailable at %s:%u", hostname, portnumber);

	return;
}

static void check_smtp_response(void *sd, char *name)
{
	char buf[1024];
	get_response_smtp(CLIENT, sd, buf, sizeof(buf));

	if (!isdigit((unsigned char)buf[0]) || (unsigned char)buf[0] > '3') {
		remcrlf(buf);
		mprintf("%s %s", ansi_warning, buf);
		mprintf(" [smtp.server] >> %s command - %s", name, ansi_failed);
		if (strncmp(name, "EHLO", 4) == 0) { /* EHLO failed. */
			say_smtp(sd, "RSET", "rset");
			say_smtp(sd, "HELO", "helo %s", hostname);
		}
		if (strncmp(name, "RSET", 4) == 0) /* RSET failed. */
			anubis_error("RSET failed. Exiting immediately.");
		else
			smtp_failed(sd);
		return;
	} else {
		if (output > NORMAL)
			mprintf(" [smtp.server] >> %s command - %s", name, ansi_ok);
		if (strncmp(name, "EHLO", 4) == 0)
			parse_ehlo_reply(sd, buf);
	}
	return;
}

/************************
  ESMTP Authentication.
*************************/

static void esmtp_auth(void *sd, char *buf)
{
	int c;
	int len;
	char *p = 0;
	char *b64buf = 0;
	char tmp[LINEBUFFER];
	memset(tmp, 0, LINEBUFFER);

	if (strstr(buf, "CRAM-MD5")) {
		unsigned char digest[16];
		static char ascii_digest[33];
		memset(digest, 0, 16);

		info(VERBOSE, "ESMTP CRAM-MD5 Authentication...");
		swrite(CLIENT, sd, "AUTH CRAM-MD5\r\n");
		get_response_smtp(CLIENT, sd, tmp, LINEBUFFER);

		if (strncmp(tmp, "334 ", 4)) { /* Server rejected AUTH */
			swrite(CLIENT, sd, "*\r\n");
			info(VERBOSE, "Server rejected the AUTH command.");
			get_response_smtp(CLIENT, sd, 0, 0);
			return;
		}

		p = strchr(tmp, ' ');
		p++;
		base64decode_string(p, &b64buf);
		#ifdef DEBUG_CODE
		info(DEBUG, "Challenge decoded: %s", b64buf);
		#endif /* DEBUG_CODE */
		cram_md5((char *)password, b64buf, digest);
		for (c = 0; c < 16; c++)
			sprintf(ascii_digest + 2 * c, "%02x", digest[c]);
		sprintf(tmp, "%s %s", username, ascii_digest);
		sprintf(tmp, "%s\r\n", base64encode_string(tmp, strlen(tmp)));
		swrite(CLIENT, sd, tmp);
		check_smtp_response(sd, "AUTH");
	}
	else if (strstr(buf, "PLAIN") && (popt & P_SSL_CLIENT)) {
		info(VERBOSE, "ESMTP PLAIN Authentication...");
		sprintf(tmp, "^%s^%s", username, password);
		len = strlen(tmp);
		for (c = len - 1; c >= 0; c--)
		{
			if (tmp[c] == '^')
				tmp[c] = '\0';
		}
		sprintf(tmp, "AUTH PLAIN %s\r\n", base64encode_string(tmp, len));
		swrite(CLIENT, sd, tmp);
		check_smtp_response(sd, "AUTH");
	}
	else if (strstr(buf, "LOGIN") && (popt & P_SSL_CLIENT)) {
		info(VERBOSE, "ESMTP LOGIN Authentication...");
		swrite(CLIENT, sd, "AUTH LOGIN\r\n");
		get_response_smtp(CLIENT, sd, tmp, LINEBUFFER);

		if (strncmp(tmp, "334 ", 4)) { /* Server rejected AUTH */
			swrite(CLIENT, sd, "*\r\n");
			info(VERBOSE, "Server rejected the AUTH command.");
			get_response_smtp(CLIENT, sd, 0, 0);
			return;
		}

		p = strchr(tmp, ' ');
		p++;
		base64decode_string(p, &b64buf);
		#ifdef DEBUG_CODE
		info(DEBUG, "Challenge decoded: %s", b64buf);
		#endif /* DEBUG_CODE */
		sprintf(tmp, "%s\r\n",
		base64encode_string((char *)username, strlen((char *)username)));
		swrite(CLIENT, sd, tmp);
		get_response_smtp(CLIENT, sd, tmp, LINEBUFFER);
		p = strchr(tmp, ' ');
		p++;
		base64decode_string(p, &b64buf);
		sprintf(tmp, "%s\r\n",
		base64encode_string((char *)password, strlen((char *)password)));
		swrite(CLIENT, sd, tmp);
		check_smtp_response(sd, "AUTH");
	}
	return;
}

/*******************
  Create an email
********************/

static void build_message(void *sd)
{
	char buf[1024];

	/*
	   REAL STEALTH MODE V2... ;-)
	*/

	if (!(popt & P_FAKENO)) {
		char tmp_mid[13];
		strncpy(tmp_mid, rand_md5(), 12);
		if (!(popt & P_FAKEHOST)) /* default */
			fakehost = allocbuf("unknown", 65);
		if (!(popt & P_FAKEIP1)) {
			sprintf(fakeip1, "%d.%d.%d.1", classc(),
			rand_se(255), rand_se(255));
		}
		if (!(popt & P_FAKEIP2)) {
			sprintf(fakeip2, "%d.%d.%d.1", classc(),
			rand_se(255), rand_se(255));
		}
		sprintf(buf,
		"Received: from unknown ([%s])\r\n\tby %s with SMTP;\r\n\t%s +0000\r\n"
		"Received: from unknown (HELO localhost) ([%s])\r\n\tby %s (8.11.6/8.11.6) "
		"with ESMTP id %s\r\n\tfor <%s>; %s +0000\r\n",
		fakeip2, (popt & P_WG1) ? ((popt & P_WG2) ? wingate2 : wingate1) : fakehost,
		get_time_string(1), fakeip1, fakeip2, tmp_mid,
		parse_address(to), get_time_string(1));

		sendline(sd, buf);
	}

	sprintf(buf,	"From: %s\r\n"
				"Reply-To: %s\r\n"
				"To: %s",
				from, (mopt & M_REPLYTO) ? reply_to : from, to);
	sendline(sd, buf);

	if (mopt & M_CC) { /* Carbon Copy */
		sprintf(buf, "Cc: %s", carbon_copy);
		sendline(sd, buf);
	}
	if (mopt & M_BCC) { /* Blind Carbon Copy */
		sprintf(buf, "Bcc: %s", blind_carbon_copy);
		sendline(sd, buf);
	}

	sprintf(buf, "Subject: %s\r\nDate: %s",
		subject, get_time_string(0));
	sendline(sd, buf);

	if (!(mopt & M_XMAIL_NO)) {
		swrite(CLIENT, sd, "X-Mailer: ");
		if (mopt & M_XMAIL_CUSTOM) {
			sendline(sd, xmailer);
		}
		else { /* default */
			swrite(CLIENT, sd, (char *)version);
			swrite(CLIENT, sd, CRLF);
		}
	}

	/*
	   Create Message-ID.
	*/

	swrite(CLIENT, sd, "Message-ID: ");
	if (mopt & M_MID_CUSTOM)
		sprintf(buf, "<%s>", messageid);
	else /* standard - random */
		sprintf(buf, "<%s>", generate_mid(hostname));
	sendline(sd, buf);

	if (mopt & M_EH1)
		sendline(sd, extra_h1);
	if (mopt & M_EH2)
		sendline(sd, extra_h2);

	sendline(sd, "MIME-Version: 1.0");
	if (mopt & M_RM) { /* Remailer Type-I support */
		remailer_cypherpunk(CLIENT, sd);
		if ((mopt & M_RMPGP) || (mopt & M_RMPGPFILE)) {
			say_smtp(sd, "EOM ", "\r\n."); /* EOM */
			return;
		}
	}

	if (mopt & M_FILEATTACH) {
		sendline(sd, "Content-Type: multipart/mixed; "
		"boundary=\"BOUNDARY\"\r\n\r\n--BOUNDARY");
	}
	sprintf(buf,	"Content-Type: text/plain; charset=%s\r\n"
				"Content-Transfer-Encoding: %s\r\n\r\n",
				(mopt & M_CHARSET) ? charset : "US-ASCII",
				(mopt & M_CHARSET) ? "8bit" : "7bit");
	sendline(sd, buf);

	/*
	   And finally send a message body.
	*/

	swrite(CLIENT, sd, msgbody);

	if (mopt & M_FILEATTACH) {
		attach_file(CLIENT, sd);
		if (!(mopt & M_MIXMASTER))
			say_smtp(sd, "EOM ", "\r\n--BOUNDARY--\r\n.");
	}
	else { /* EOM (End Of Message) */
		if (!(mopt & M_MIXMASTER))
			say_smtp(sd, "EOM ", ".");
	}
	return;
}

static void attach_file(int method, void *sd)
{
	char buf[1024];

	sprintf(buf, "\r\n--BOUNDARY\r\n"
		"Content-Type: %s; name=\"%s\"\r\n"
		"Content-Disposition: attachment; filename=\"%s\"\r\n"
		"Content-Transfer-Encoding: %s\r\n\r\n", examine_file_type(filename),
		filename, filename, trans_ptr == uu_base64 ? "base64" : "x-uue");
	sendline(sd, buf);

	if (trans_ptr == uu_std) { /* Uuencode */
		sprintf(buf, "begin %o %s", uu_mode, filename);
		sendline(sd, buf);
	}
	encode_mime(method, sd); /* Uuencode & Base64 encoding */

	if (trans_ptr == uu_std)
		sendline(sd, "end");
	return;
}

/***************************
  Remailer Type-I support
****************************/

static void remailer_cypherpunk(int method, void *sd)
{
	char buf[1024];

	if (!(mopt & M_RMPGP) && !(mopt & M_RMPGPFILE)) {
		if (mopt & M_RMRRT) {
			sprintf(buf, "\r\n::\r\nAnon-To: %s", rm_rrt);
			sendline(sd, buf);
		}
		else if (mopt & M_RMPOST) {
			sprintf(buf, "\r\n::\r\nAnon-Post-To: %s", rm_post);
			sendline(sd, buf);
		}
		if ((mopt & M_RMLT) || (mopt & M_RMRLT)) {
			sprintf(buf, "Latent-Time: +%s%s", rm_lt, (mopt & M_RMRLT) ? "r" : "");
			sendline(sd, buf);
		}
		sendline(sd, "\r\n##");
		if (mopt & M_RMHEADER)
			sendline(sd, rm_header);
	}
	else {
		sendline(sd, "\r\n::\r\nEncrypted: PGP\r\n\r\n");

		#ifndef NOGPG
		if (mopt & M_RMPGP) {
			sendline(sd, msgbody);
		}
		else if (mopt & M_RMPGPFILE)
		#endif /* not NOGPG */
			attach_pgp_file(method, sd);
	}
	return;
}

/****************************
  Remailer Type-II support
*****************************/

static void remailer_mixmaster(void)
{
	int n;
	int local;
	char *def_args[3];
	char *home;
	char *mixmaster_dir = (char *)DEFAULT_MIXMASTER_PATH;

	info(NORMAL, "Redirecting to the Mixmaster...");

	home = get_homedir(local_user);
	if (home == 0)
		home = (char *)"";

	n = strlen(home) + strlen(mixmaster_dir) + 2;
	n = n > MAXPATHLEN ? MAXPATHLEN : n + 1;
	execpath = (char *)xmalloc(n);
	#ifdef HAVE_SNPRINTF
	snprintf(execpath, n - 1,
	#else
	sprintf(execpath,
	#endif /* HAVE_SNPRINTF */
	"%s/%s", home, mixmaster_dir);

	def_args[0] = strrchr(execpath, '/') + 1;
	def_args[1] = "-m";
	def_args[2] = 0;
	execargs = def_args;

	local = make_local_connection(execpath, execargs);
	if (local == -1)
		quit(EXIT_FAILURE);

	build_message((void *)local);
	close_socket(local);
	return;
}

static void smtp_failed(void *sd)
{
	say_smtp(sd, "RSET", "rset");
	say_smtp(sd, "QUIT", "quit");
	anubis_error("Connection terminated.");
	return;
}

static void smtp_success(void)
{
	info(VERBOSE, "%ld bytes sent.", bytes_sent);
	info(NORMAL, "Connection closed succesfully.");
	return;
}

#endif /* not TUNNEL_OPTIMIZE */

/* EOF */

