/*
    ssl.c

    This file is a part of Anubis.
    Copyright (C) 2001, 2002 The Anubis Team.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
   OpenSSL package support.
*/

#include "headers.h"
#include "extern.h"

#ifndef NOSSL
#define X509BUFSIZE 4096

static void ssl_error(char *);
static void cipher_info(SSL *);

void init_ssl_libs(void)
{
	SSL_library_init();
	SSL_load_error_strings();

	if (!(RAND_status())) {
		char tmp[33];
		info(VERBOSE, "Seeding random number generator...");
		safe_strcpy(tmp, rand_md5());
		RAND_seed(tmp, strlen(tmp));
		if (!(RAND_status()))
			info(VERBOSE, "Unable to seed random number generator.");
	}
	return;
}

static void ssl_error(char *txt)
{
	char string_error[256];
	memset(string_error, 0, sizeof(string_error));
	ERR_error_string(ERR_get_error(), string_error);

	if (output != SILENT) {
		#ifndef NOLOGS
		if ((popt & P_DAEMON) && !(popt & P_FOREGROUND))
			syslog(LOG_ERR | LOG_MAIL, string_error);
		else
		#endif /* not NOLOGS */
			mprintf("%s>%s %s", ansi_red, ansi_normal, string_error);
	}
	anubis_error(txt);
	return;
}

SSL_CTX *init_ssl_client(void)
{
	SSL_CTX *ctx_local = 0;
	SSL_METHOD *method = 0;

	if (!(method = SSLv23_client_method())) {
		ssl_error("SSLv23_client_method() failed.");
		return 0;
	}
	if (!(ctx_local = SSL_CTX_new(method))) {
		ssl_error("Can't create SSL_CTX object.");
		return 0;
	}
	SSL_CTX_set_default_verify_paths(ctx_local);
	if (!(SSL_CTX_set_cipher_list(ctx_local, "DEFAULT"))) {
		ssl_error("SSL_CTX_set_cipher_list() failed.");
		return 0;
	}
	SSL_CTX_set_verify(ctx_local, SSL_VERIFY_NONE, 0);

	return (SSL_CTX *)ctx_local;
}

SSL *start_ssl_client(int sd_server, SSL_CTX *ctx_ptr)
{
	SSL *ssl_local = 0;
	info(VERBOSE, "Initializing TLS/SSL connection with server...");

	if (!(ssl_local = SSL_new(ctx_ptr))) {
		ssl_error("Can't create a new SSL structure for a connection.");
		return 0;
	}
	if (!(SSL_set_fd(ssl_local, sd_server))) {
		ssl_error("SSL_set_fd() failed.");
		return 0;
	}
	SSL_set_connect_state(ssl_local);
	if (SSL_connect(ssl_local) <= 0) {
		ssl_error("TLS/SSL handshake error !");
		return 0;
	}
	popt |= P_SSL_CLIENT;

	if (output > NORMAL)
		cipher_info(ssl_local);

	return (SSL *)ssl_local;
}

/**************************
  TLS/SSL SERVER support
***************************/

SSL_CTX *init_ssl_server(void)
{
	SSL_CTX *ctx_local = 0;
	SSL_METHOD *method = 0;

	check_file_mode(ssl_cert);
	check_file_mode(ssl_key);

	if (!(method = SSLv23_server_method())) {
		ssl_error("SSLv23_server_method() failed.");
		return 0;
	}
	if (!(ctx_local = SSL_CTX_new(method))) {
		ssl_error("Can't create SSL_CTX object.");
		return 0;
	}
	ssl_cert = ssl_cert ? ssl_cert : (char *)DEFAULT_SSL_PEM;
	if (SSL_CTX_use_certificate_file(ctx_local, ssl_cert, SSL_FILETYPE_PEM) <= 0) {
		ssl_error("SSL_CTX_use_certificate_file() failed.");
		return 0;
	}
	ssl_key = ssl_key ? ssl_key : ssl_cert;
	if (SSL_CTX_use_PrivateKey_file(ctx_local, ssl_key, SSL_FILETYPE_PEM) <= 0) {
		ssl_error("SSL_CTX_use_PrivateKey_file() failed.");
		return 0;
	}
	if (!(SSL_CTX_check_private_key(ctx_local))) {
		ssl_error("Private key does not match the certificate public key.");
		return 0;
	}
	if (!(SSL_CTX_set_cipher_list(ctx_local, "DEFAULT"))) {
		ssl_error("SSL_CTX_set_cipher_list() failed.");
		return 0;
	}
	return (SSL_CTX *)ctx_local;
}

SSL *start_ssl_server(int sd_client, SSL_CTX *ctx_ptr)
{
	SSL *ssl_local = 0;
	info(VERBOSE, "Initializing TLS/SSL connection with a client...");

	if (!(ssl_local = SSL_new(ctx_ptr))) {
		ssl_error("Can't create a new SSL structure for a connection.");
		return 0;
	}
	if (!(SSL_set_fd(ssl_local, sd_client))) {
		ssl_error("SSL_set_fd() failed.");
		return 0;
	}
	SSL_set_accept_state(ssl_local);
	if (SSL_accept(ssl_local) <= 0) {
		ssl_error("TLS/SSL handshake error !");
		return 0;
	}
	popt |= P_SSL_SERVER;

	if (output > NORMAL)
		cipher_info(ssl_local);

	return (SSL *)ssl_local;
}

/***********************
   Certificate info...
************************/

static void cipher_info(SSL *ssl_local)
{
	X509 *cert = 0;
	char *x509buf = 0;
	int bits;

	SSL_CIPHER *cipher = SSL_get_current_cipher(ssl_local);

	if (cipher) {
		SSL_CIPHER_get_bits(cipher, &bits);
		info(VERBOSE, "%s connection using %s (%u bits)",
		SSL_CIPHER_get_version(cipher), SSL_CIPHER_get_name(cipher), bits);
	}
	cert = SSL_get_peer_certificate(ssl_local);
	if (cert == 0)
		return;

	info(VERBOSE, "Server public key is %d bits",
	EVP_PKEY_bits(X509_get_pubkey(cert)));

	x509buf = (char *)xmalloc(X509BUFSIZE);

	info(VERBOSE, "Certificate:");
	if (!(X509_NAME_oneline(X509_get_subject_name(cert), x509buf, X509BUFSIZE))) {
		ssl_error("X509_NAME_oneline [subject] error !");
		return;
	}
	info(VERBOSE, "subject: %s", x509buf);
	if (!(X509_NAME_oneline(X509_get_issuer_name(cert), x509buf, X509BUFSIZE))) {
		ssl_error("X509_NAME_oneline [issuer] error !");
		return;
	}
	info(VERBOSE, "issuer:  %s", x509buf);

	/*
	   Free unused memory.
	*/

	free(x509buf);
	if (cert)
		X509_free(cert);
	return;
}

void end_ssl(int method, SSL *ssl_ptr, SSL_CTX *ctx_ptr)
{
	if ((popt & P_SSL_CLIENT) || (popt & P_SSL_SERVER)) {
		if (ssl_ptr) {
			SSL_shutdown(ssl_ptr);
			SSL_free(ssl_ptr);
		}
		if (ctx_ptr)
			SSL_CTX_free(ctx_ptr);

		if (method == CLIENT)
			popt &= ~P_SSL_CLIENT;
		else if (method == SERVER)
			popt &= ~P_SSL_SERVER;
	}
	return;
}

#endif /* not NOSSL */

/* EOF */

