/*
    tunnel.c

    This file is a part of Anubis.
    Copyright (C) 2001, 2002 The Anubis Team.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "headers.h"
#include "extern.h"

#define MAX_CLIENTS 50

static void standard_inout(void);
static void daemon_loop(int);
static void transfer_message(void *, void *);
static int  transfer_command(void *, void *, char *);
static void process_command(void *, void *, char *, int);
static void transfer_header(void *, void *);
static void process_header(char *);
static void transfer_body(void *, void *);
static void static_body_transfer(void *, void *);
static void dynamic_body_transfer(void *, void *);
static void add_remailer_commands(void *);
#ifndef NOAUTH
static int  auth_ident(struct sockaddr_in *);
static void tunnel_auth(void);
#endif /* not NOAUTH */

#ifndef NOFORK
static void daemonize(void)
{
	pid_t pid;

	chdir("/");
	umask(0);

	pid = fork();
	if (pid > 0) /* parent process */
		quit(0);
	else if (pid == -1) /* fork() failed */
		anubis_error("Can't fork.");

	if (setsid() == -1)
		anubis_error("setsid() failed.");

	signal(SIGHUP, SIG_IGN);
	close(0);
	close(1);
	close(2);
	popt |= P_DAEMON;

	#ifndef NOLOGS
	openlog("anubis", LOG_PID, 0);
	syslog(LOG_INFO, "%s daemon startup succeeded.", version);
	#endif /* not NOLOGS */

	return;
}

static void sig_cld(int code)
{
	pid_t pid;
	int status;

	while ((pid = waitpid(-1, &status, WNOHANG)) > 0)
	{
		nchild--;
		info(VERBOSE, "[%d] finished. Exit status: %s. %d client(s) left.",
		pid, WIFEXITED(status) ? "OK" : "ERROR", nchild);
	}
	return;
}
#endif /* not NOFORK */

void smtp_tunnel_init(char *host, unsigned int port)
{
	if (topt & T_STD_INOUT) { /* stdin/stdout */
		if (output != DEBUG)
			output = SILENT;
		popt |= P_FOREGROUND;
		popt |= P_SSL_NO;
		standard_inout();
	}
	else {
		int sd_bind = bind_and_listen(host, port);
		#ifndef NOFORK
		if (!(popt & P_FOREGROUND_INIT))
			daemonize();
		else
			popt |= P_FOREGROUND;
		signal(SIGCHLD, sig_cld);
		#else
			popt |= P_FOREGROUND;
		#endif /* not NOFORK */
		daemon_loop(sd_bind);
	}
	return;
}

static void service_unavailable(int sd_client)
{
	char buf[256];

	sprintf(buf, "421 %s Service not available, "
	"closing transmission channel.\r\n", hostname);
	swrite(SERVER, (void *)sd_client, buf);

	popt &= ~P_ERROR;
	return;
}

/**********************************************
  Run an outgoing mail processor on standard
  input and output as described in RFC 821.
***********************************************/

static void standard_inout(void)
{
	int sd_client = -1; /* stdin/stdout */
	int sd_server = 0;

	/*
	   Read the config file (the Tunnel's owner).
	*/

	if (!(popt & P_NORC)) {
		open_config_file(LOCAL, local_user);
		read_config_file();
		if (fp_config)
			fp_config = (FILE *)fclose(fp_config);
	}
	#if !defined(NOREGEX) || defined(PCRE)
	regex_position = 0;
	#endif /* not NOREGEX or PCRE */

	safe_strcpy(remote_user, anubis_getlogin());
	tunnel_auth();

	if (topt & T_LOCAL_MTA) {
		sd_server = make_local_connection(execpath, execargs);
		if (sd_server == -1) {
			service_unavailable(sd_client);
			close_socket(sd_client);
			quit(EXIT_FAILURE);
		}
	}
	else {
		sd_server = make_remote_connection(hostname, portnumber);
		if (sd_server == -1) {
			service_unavailable(sd_client);
			close_socket(sd_client);
			quit(EXIT_FAILURE);
		}
	}
	remote_client = (void *)sd_client;
	remote_server = (void *)sd_server;
	transfer_message((void *)sd_client, (void *)sd_server);

	#if !defined(NOREGEX) || defined(PCRE)
	if (fp_config)
		fp_config = (FILE *)fclose(fp_config);
	#endif /* not NOREGEX or PCRE */

	close_socket(sd_server);
	return;
}

/*********************
  Anubis DAEMON loop
**********************/

static void daemon_loop(int sd_bind)
{
	struct sockaddr_in addr;
	struct sockaddr_in laddr;
	struct hostent *hp = 0;
	int sd_client = 0;
	int sd_server = 0;
	#ifndef NOFORK
	int childpid = 0;
	#endif /* not NOFORK */
	socklen_t addrlen;

	addrlen = sizeof(addr);
	info(NORMAL, "Anubis SMTP Tunnel is running...");

	for (;;)
	{
		sd_client = accept(sd_bind, (struct sockaddr *)&addr, &addrlen);
		if (sd_client < 0) {
			if (errno == EINTR)
				continue;
			else
				anubis_error("accept() failed. %s", strerror(errno));
		}
		if (!(topt & T_LOCAL_MTA)) {
			if ((hp = gethostbyname(hostname)) == 0) {
				laddr.sin_addr.s_addr = inet_addr(hostname);
				if ((int)laddr.sin_addr.s_addr == -1)
					anubis_error("Unknown host %s", hostname);
			}
			else
				memcpy((char *)&laddr.sin_addr.s_addr, hp->h_addr, hp->h_length);
		}

		/*
		   Read the config file (the Tunnel's owner).
		*/

		if (!(popt & P_NORC)) {
			open_config_file(LOCAL, local_user);
			read_config_file();
			if (fp_config)
				fp_config = (FILE *)fclose(fp_config);
		}
		#if !defined(NOREGEX) || defined(PCRE)
		regex_position = 0;
		#endif /* not NOREGEX or PCRE */

		nchild++;
		if (nchild > MAX_CLIENTS) {
			info(NORMAL, "Too many clients. Connection from %s:%u rejected.",
			inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));
			close_socket(sd_client);
			nchild--;
		}
		else if (strncmp(inet_ntoa(laddr.sin_addr), "127.0.0.1", 9) == 0
			&& tunnel_portnumber == portnumber && !(topt & T_LOCAL_MTA)) {
			info(NORMAL, "Loop not allowed. Connection rejected.");
			close_socket(sd_client);
			nchild--;
		}
		else {
			info(NORMAL, "Connected from %s:%u",
			inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));

			#ifndef __BEOS__
			setsockopt(sd_client, SOL_SOCKET, SO_LINGER, 0, 0);
			setsockopt(sd_client, SOL_SOCKET, SO_KEEPALIVE, 0, 0);
			#endif /* not __BEOS__ */

			#ifndef NOFORK
			if ((childpid = fork()) < 0)
				anubis_error("daemon: can't fork.");
			else if (childpid == 0) { /* child process */
			#endif /* not NOFORK */

				if (strncmp(inet_ntoa(addr.sin_addr), "127.0.0.1", 9) == 0) {
					topt |= T_LOCAL_CONNECTION; /* connection from a local machine */
					#ifndef NOAUTH
					if (auth_ident(&addr) == -1) /* IDENT auth. */
						popt &= ~P_ERROR;
					#endif /* not NOAUTH */
				}

				#ifndef NOFORK
				signal(SIGCHLD, SIG_IGN);

				if (topt & T_LOCAL_MTA) {
					sd_server = make_local_connection(execpath, execargs);
					if (sd_server == -1) {
						service_unavailable(sd_client);
						close_socket(sd_client);
						#ifndef NOFORK
						quit(EXIT_FAILURE);
						#else
						continue;
						#endif /* not NOFORK */
					}
				} else {
				#endif /* not NOFORK */
					sd_server = make_remote_connection(hostname, portnumber);
					if (sd_server == -1) {
						service_unavailable(sd_client);
						close_socket(sd_client);
						#ifndef NOFORK
						quit(EXIT_FAILURE);
						#else
						continue;
						#endif /* not NOFORK */
					}
				#ifndef NOFORK
				}
				#endif /* not NOFORK */

				remote_client = (void *)sd_client;
				remote_server = (void *)sd_server;
				transfer_message((void *)sd_client, (void *)sd_server);

				#if !defined(NOREGEX) || defined(PCRE)
				if (fp_config)
					fp_config = (FILE *)fclose(fp_config);
				#endif /* not NOREGEX or PCRE */

				#ifndef NOSSL
				end_ssl(SERVER, ssl_server, ctx_server);
				end_ssl(CLIENT, ssl_client, ctx_client);
				ssl_server = 0;
				ssl_client = 0;
				#endif /* not NOSSL */
				close_socket(sd_server);
				close_socket(sd_client);
				if (popt & P_ERROR)
					info(NORMAL, "Connection terminated.");
				else
					info(NORMAL, "Connection closed succesfully.");
			#ifndef NOFORK
				quit(0);
			 }
			 close_socket(sd_client);
			#else
			 nchild--;
			 free_mem();
			 mopt = 0;
			 topt &= ~T_LOCAL_CONNECTION;
			 topt &= ~T_STARTTLS;
			 topt &= ~T_BOUNDARY;
			 topt &= ~T_AUTH_LOCAL;
			 topt &= ~T_IDENT_ONLY;
			 mopt &= ~M_SUBJECT;
			 mopt &= ~M_BODY;
			 popt &= ~P_ERROR;
			#endif /* not NOFORK */
		}
	}
	return;
}

/*******************************
  Regular Expressions support
********************************/

#ifndef NOREGEX
int regex_match(char *regex, char *line)
{
	int ret;
	regex_t preg;

	if (regcomp(&preg, regex, REG_EXTENDED | REG_NOSUB)) {
		anubis_error("regcomp() failed. Bad REGEX gramma.");
		popt &= ~P_ERROR;
		return 0;
	}
	ret = regexec(&preg, line, 0, 0, 0);
	regfree(&preg);

	if (ret == REG_NOMATCH)
		return 0;
	else
		return 1;
}
#endif /* not NOREGEX */

#ifdef PCRE
int regex_match(char *regex, char *line)
{
	pcre *re;
	const char *error;
	int error_offset;
	int rc;

	re = pcre_compile(regex, 0, &error, &error_offset, 0);
	if (re == 0) {
		anubis_error("pcre_compile() failed at offset %d: %s.", error_offset, error);
		popt &= ~P_ERROR;
		return 0;
	}
	rc = pcre_exec(re, 0, line, strlen(line), 0, 0, 0, 0);

	if (rc < 0) {
		switch(rc)
		{
			case PCRE_ERROR_NOMATCH:
				break;
			default:
				break;
		}
		return 0;
	}
	return 1;
}
#endif /* not PCRE */

/******************
  The Tunnel core
*******************/

static void transfer_message(void *sd_client, void *sd_server)
{
	char command[LINEBUFFER];

	/*
	   First of all transfer a welcome message.
	*/

	info(NORMAL, "Transferring a message...");
	get_response_smtp(CLIENT, sd_server, command, LINEBUFFER);

	if (strncmp(command, "220 ", 4) == 0 && strstr(command, version) == 0) {
		char *ptr = 0;
		char *banner_ptr = 0;
		char host[65];
		char banner_backup[LINEBUFFER];
		safe_strcpy(banner_backup, command);

		if ((banner_ptr = strchr(banner_backup, ' '))) {
			banner_ptr++;
			safe_strcpy(host, banner_ptr);
			if ((ptr = strchr(host, ' ')))
				*ptr = '\0';
			do {
				banner_ptr++;
			} while (*banner_ptr != ' ');
			banner_ptr++;
			#ifdef HAVE_SNPRINTF
			snprintf(command, LINEBUFFER,
			#else
			sprintf(command,
			#endif /* HAVE_SNPRINTF */
			"220 %s (%s) %s", host, version, banner_ptr);
		}
	}
	swrite(SERVER, sd_client, command);

	/*
	   Then process the commands...
	*/

	while (recvline(SERVER, sd_client, command, LINEBUFFER))
	{
		if (topt & T_STD_INOUT) {
			remcrlf(command);
			strncat(command, CRLF, 2);
		}
		#ifdef DEBUG_CODE
		if (output == DEBUG)
			fprintf(stderr, "CLIENT: %s", command);
		#endif /* DEBUG_CODE */

		process_command(sd_client, sd_server, command, LINEBUFFER);
		sd_client = remote_client;
		sd_server = remote_server;

		if (popt & P_ERROR)
			break;
		if (strlen(command) == 0)
			continue;

		if (transfer_command(sd_client, sd_server, command) == 0)
			break;
	}
}

static void process_command(void *sd_client, void *sd_server, char *command, int size)
{
	char buf[LINEBUFFER];
	safe_strcpy(buf, command); /* make a back-up */
	change_to_lower(buf);

	/*
	   The Tunnel auth.
	*/

	#ifndef NOAUTH
	if ((topt & T_LOCAL_CONNECTION) && (topt & T_AUTH_LOCAL) && !(topt & T_IDENT_ONLY)) {
		if (strncmp(buf, "mail from:", 10) == 0) {
			char *ptr = 0;
			safe_strcpy(remote_user, parse_address(command));
			if ((ptr = strstr(remote_user, "@")))
				*ptr = '\0';
			tunnel_auth();
		}
	}
	#endif /* not NOAUTH */

	#if !defined(NOREGEX) || defined(PCRE)
	if (regex_position) {
		char regex[LINEBUFFER];
		fseek(fp_config, regex_position, SEEK_SET);
		while (read_regex_block(COMMAND, regex, LINEBUFFER) != 0)
		{
			if (regex_match(regex, command)) {
				while (read_action_block(command) != 0)
				{
					safe_strcpy(buf, command);
				}
			}
		}
	}
	#endif /* not NOREGEX or PCRE */

	if (strncmp(buf, "starttls", 8) == 0) {
		#ifndef NOSSL
		char reply[LINEBUFFER];

		if (!(topt & T_STARTTLS) || (popt & P_SSL_NO)) {
			swrite(SERVER, sd_client, "503 5.5.0 TLS not available\r\n");
			strncpy(command, "", 1);
			return;
		}

		/*
		   Make the TLS/SSL connection with ESMTP server.
		*/

		if (!(topt & T_LOCAL_MTA)) {
			swrite(CLIENT, sd_server, "starttls\r\n");
			get_response_smtp(CLIENT, sd_server, reply, LINEBUFFER);

			if (!isdigit((unsigned char)reply[0]) || (unsigned char)reply[0] > '3') {
				remcrlf(reply);
				mprintf("%s %s", ansi_warning, reply);
				anubis_error("STARTTLS command failed.");
				return;
			}
			ctx_client = init_ssl_client();
			if (popt & P_ERROR)
				return;
			ssl_client = start_ssl_client((int)sd_server, ctx_client);
			if (popt & P_ERROR)
				return;
			sd_server = remote_server = (void *)ssl_client;
		}

		/*
		   Make the TLS/SSL connection with SMTP client
		   (client connected with the Tunnel).
		*/

		ctx_server = init_ssl_server();
		if (popt & P_ERROR) {
			swrite(SERVER, sd_client, "454 4.3.3 TLS not available\r\n");
			return;
		}
		swrite(SERVER, sd_client, "220 2.0.0 Ready to start TLS\r\n");
		ssl_server = start_ssl_server((int)sd_client, ctx_server);
		if (popt & P_ERROR)
			return;
		sd_client = remote_client = (void *)ssl_server;

		#else
		swrite(SERVER, sd_client, "503 5.5.0 TLS not available\r\n");
		#endif /* not NOSSL */

		strncpy(command, "", 1);
		return;
	}
	strncpy(command, buf, size);
	return;
}

static int transfer_command(void *sd_client, void *sd_server, char *command)
{
	char reply[LINEBUFFER];
	char buf[LINEBUFFER];

	safe_strcpy(buf, command);
	change_to_lower(buf);
	swrite(CLIENT, sd_server, command);

	if (strncmp(buf, "ehlo", 4) == 0) {
		get_response_smtp(CLIENT, sd_server, reply, LINEBUFFER);

		if (strstr(reply, "STARTTLS"))
			topt |= T_STARTTLS; /* Yes, we can use the TLS/SSL encryption. */

		/*
		   Remove the STARTTLS command from the EHLO list
		   if NOSSL or --nossl is specified or SSL is not available.
		*/

		if ((topt & T_STARTTLS) && (popt & P_SSL_NO)) {
			char *starttls1 = "250-STARTTLS";
			char *starttls2 = "STARTTLS";
			if (strstr(reply, starttls1))
				remline(reply, starttls1);
			else if (strstr(reply, starttls2))
				remline(reply, starttls2);
		}
	}
	else
		get_response_smtp(CLIENT, sd_server, reply, LINEBUFFER);

	swrite(SERVER, sd_client, reply);

	if (strncmp(buf, "quit", 4) == 0)
		return 0; /* The QUIT command */
	else if (strncmp(buf, "data", 4) == 0) {
		if (isdigit((unsigned char)reply[0]) && (unsigned char)reply[0] < '4') {
			transfer_header(sd_client, sd_server);
			transfer_body(sd_client, sd_server);
		}
	}
	return 1; /* OK */
}

static void transfer_header(void *sd_client, void *sd_server)
{
	char *header_buf = 0;
	char header_line[LINEBUFFER];

	#if !defined(NOREGEX) || defined(PCRE)
	addlist = (char *)xmalloc(1);
	remlist = (char *)xmalloc(1);
	#endif /* not NOREGEX or PCRE */

	header_buf = (char *)xmalloc(1);
	while (recvline(SERVER, sd_client, header_line, LINEBUFFER))
	{
		if (topt & T_STD_INOUT) {
			remcrlf(header_line);
			strncat(header_line, CRLF, 2);
		}
		process_header(header_line);
		header_buf = (char *)xrealloc((char *)header_buf,
		strlen(header_buf) + strlen(header_line) + 1);
		strncat(header_buf, header_line, strlen(header_line));

		if (strncmp(header_line, CRLF, 2) == 0) {
			#if !defined(NOREGEX) || defined(PCRE)
			if (remlist) {
				if (strlen(remlist) != 0) {
					char *pos = 0;
					char oneline[LINEBUFFER];
					char pline[LINEBUFFER];
					do
					{
						getoneline(remlist, oneline, LINEBUFFER);
						remcrlf(oneline);
						remline(remlist, oneline);
						pos = header_buf;
						for (;;)
						{
							getoneline(pos, pline, LINEBUFFER);
							remcrlf(pline);
							if (strlen(pline) <= 2)
								break;
							pos = strchr(pos, '\n');
							pos++;
							if (regex_match(oneline, pline))
								remline(header_buf, pline);
						}
					} while (strlen(remlist) != 0);
				}
				xfree(remlist);
			}
			if (addlist) {
				if (strlen(addlist) != 0)
					remcrlf(header_buf);
			}
			#endif /* not NOREGEX or PCRE */

			swrite(CLIENT, sd_server, header_buf);
			if (header_buf)
				free(header_buf);

			#if !defined(NOREGEX) || defined(PCRE)
			if (addlist) {
				if (strlen(addlist) != 0) {
					swrite(CLIENT, sd_server, addlist);
					swrite(CLIENT, sd_server, CRLF);
				}
				xfree(addlist);
			}
			#endif /* not NOREGEX or PCRE */

			return;
		}
	}
}

static void process_header(char *header_line)
{
	#if !defined(NOREGEX) || defined(PCRE)
	if (regex_position) {
		char regex[LINEBUFFER];
		fseek(fp_config, regex_position, SEEK_SET);
		while (read_regex_block(HEADER, regex, LINEBUFFER) != 0)
		{
			if (regex_match(regex, header_line)) {
				while (read_action_block(header_line) != 0)
				{ }
			}
		}
	}
	#endif /* not NOREGEX or PCRE */

	if (strncmp(header_line, "Subject:", 8) == 0) {
		subject = allocbuf(header_line, 512);
		remcrlf(subject);
		mopt |= M_SUBJECT;
	}

	/*
	   If there are any attachments, find the BOUNDARY.
	*/

	if (strncmp(header_line, "Content-Type:", 13) == 0) {
		char *ptr1 = 0;
		char *ptr2 = 0;
		char buf[LINEBUFFER];
		safe_strcpy(buf, header_line);
		ptr1 = strstr(buf, "boundary");
		if (ptr1 == 0) {
			ptr1 = strstr(buf, "BOUNDARY");
			if (ptr1 == 0)
				return;
		}
		topt |= T_BOUNDARY;
		ptr2 = parse_line_option(ptr1);
		boundary = (char *)xmalloc(strlen(ptr2) + 3);
		if (*ptr2 == '"') {
			ptr2++;
			sprintf(boundary, "--%s", ptr2);
			ptr2 = strstr(boundary, "\"");
			*ptr2 = '\0';
		}
		else
			sprintf(boundary, "--%s", ptr2);
	}
	return;
}

static void transfer_body(void *sd_client, void *sd_server)
{
	if ((mopt & M_PGP) || (mopt & M_PGPSIGN) || (mopt & M_RM) || (mopt & M_BODYATTACH)
	|| (mopt & M_SIGNATURE) || (mopt & M_ROT13) || (mopt & M_ROT13B))
		static_body_transfer(sd_client, sd_server);
	else
		dynamic_body_transfer(sd_client, sd_server);
	return;
}

static void static_body_transfer(void *sd_client, void *sd_server)
{
	int nb = 0;
	char body_line[LINEBUFFER];

	msgbody = (char *)xmalloc(1);
	mopt |= M_BODY;

	/*
	   If there are some attachments...
	*/

	if (topt & T_BOUNDARY) {
		nb = strlen(boundary);
		while (recvline(SERVER, sd_client, body_line, LINEBUFFER))
		{
			if (topt & T_STD_INOUT) {
				remcrlf(body_line);
				strncat(body_line, CRLF, 2);
			}
			swrite(CLIENT, sd_server, body_line);
			if (strncmp(body_line, boundary, nb) == 0) {
				while (recvline(SERVER, sd_client, body_line, LINEBUFFER))
				{
					if (topt & T_STD_INOUT) {
						remcrlf(body_line);
						strncat(body_line, CRLF, 2);
					}
					swrite(CLIENT, sd_server, body_line);
					if (strncmp(body_line, CRLF, 2) == 0)
						break;
				}
				break;
			}
		}

		/*
		   Now we reached the message body...
		*/

		while (recvline(SERVER, sd_client, body_line, LINEBUFFER))
		{
			if (topt & T_STD_INOUT) {
				remcrlf(body_line);
				strncat(body_line, CRLF, 2);
			}
			if (strncmp(body_line, boundary, nb) == 0) {
				remcrlf(msgbody);
				remcrlf(msgbody);

				if (mopt & M_RM)
					add_remailer_commands(sd_server);

				check_all_files(remote_user);
				check_rot13();
				#ifndef NOGPG
				check_pgp();
				#endif /* not NOGPG */

				popt &= ~P_ERROR;

				swrite(CLIENT, sd_server, msgbody);
				#ifndef NOGPG
				if ((mopt & M_PGP) || (mopt & M_PGPSIGN))
					swrite(CLIENT, sd_server, CRLF);
				#endif /* not NOGPG */
				swrite(CLIENT, sd_server, body_line);
				break;
			}
			msgbody = (char *)xrealloc((char *)msgbody,
			strlen(msgbody) + strlen(body_line) + 1);
			strncat(msgbody, body_line, strlen(body_line));
		}

		/*
		   Transfer everything else...
		*/

		dynamic_body_transfer(sd_client, sd_server);
	}

	/*
	   else... No attachments.
	*/

	else {
		while (recvline(SERVER, sd_client, body_line, LINEBUFFER))
		{
			if (topt & T_STD_INOUT) {
				remcrlf(body_line);
				strncat(body_line, CRLF, 2);
			}
			if (strncmp(body_line, "."CRLF, 3) == 0) { /* EOM */
				remcrlf(msgbody);

				if (mopt & M_RM)
					add_remailer_commands(sd_server);

				check_all_files(remote_user);
				check_rot13();
				#ifndef NOGPG
				check_pgp();
				#endif /* not NOGPG */

				popt &= ~P_ERROR;

				swrite(CLIENT, sd_server, msgbody);
				swrite(CLIENT, sd_server, "."CRLF);
				recvline(CLIENT, sd_server, body_line, LINEBUFFER);
				swrite(SERVER, sd_client, body_line);
				break;
			}
			msgbody = (char *)xrealloc((char *)msgbody,
			strlen(msgbody) + strlen(body_line) + 1);
			strncat(msgbody, body_line, strlen(body_line));
		}
	}
	return;
}

static void dynamic_body_transfer(void *sd_client, void *sd_server)
{
	char body_line[LINEBUFFER];

	while (recvline(SERVER, sd_client, body_line, LINEBUFFER))
	{
		if (topt & T_STD_INOUT) {
			remcrlf(body_line);
			strncat(body_line, CRLF, 2);
		}
		swrite(CLIENT, sd_server, body_line);
		if (strncmp(body_line, "."CRLF, 3) == 0) { /* EOM */
			recvline(CLIENT, sd_server, body_line, LINEBUFFER);
			swrite(SERVER, sd_client, body_line);
			return;
		}
	}
}

static void add_remailer_commands(void *sd_server)
{
	char buf[1024];

	if (!(mopt & M_RMPGP)) {
		if (mopt & M_RMRRT) {
			sprintf(buf, "::\r\nAnon-To: %s", rm_rrt);
			sendline(sd_server, buf);
		}
		else if (mopt & M_RMPOST) {
			sprintf(buf, "::\r\nAnon-Post-To: %s", rm_post);
			sendline(sd_server, buf);
		}
		if ((mopt & M_RMLT) || (mopt & M_RMRLT)) {
			sprintf(buf, "Latent-Time: +%s%s", rm_lt, (mopt & M_RMRLT) ? "r" : "");
			sendline(sd_server, buf);
		}
		sendline(sd_server, CRLF);
		if (mopt & M_RMHEADER) {
			sendline(sd_server, "##");
			sendline(sd_server, rm_header);
		}
	}
	#ifndef NOGPG
	else
		sendline(sd_server, "::\r\nEncrypted: PGP\r\n\r\n");
	#endif /* not NOGPG */

	return;
}

/**************************
  IDENT protocol support
***************************/

#ifndef NOAUTH
static int auth_ident(struct sockaddr_in *addr)
{
	struct sockaddr_in ident;
	char buf[LINEBUFFER], user[65];
	int sd = 0;

	if ((sd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		anubis_error("IDENT: socket() failed. %s", strerror(errno));
		return -1;
	}
	memcpy(&ident, addr, sizeof(ident));
	ident.sin_port = htons(113); /* IDENT port number */

	if (connect(sd, (struct sockaddr *)&ident, sizeof(ident)) < 0) {
		anubis_error("IDENT: connect() failed. %s", strerror(errno));
		close_socket(sd);
		return -1;
	}

	#ifdef HAVE_SNPRINTF
	snprintf(buf, LINEBUFFER,
	#else
	sprintf(buf,
	#endif /* HAVE_SNPRINTF */
	"%u , %u\r\n", ntohs(addr->sin_port), tunnel_portnumber);

	if (send(sd, buf, strlen(buf), 0) == -1) {
		anubis_error("IDENT: send() failed. %s", strerror(errno));
		close_socket(sd);
		return -1;
	}
	if (recvline(CLIENT, (void *)sd, buf, LINEBUFFER) == -1) {
		anubis_error("IDENT: recvline() failed. %s", strerror(errno));
		close_socket(sd);
		return -1;
	}
	close_socket(sd);

	if (sscanf(buf, "%*[^:]: USERID :%*[^:]:%s", user) != 1) {
		info(VERBOSE, "Incorrect data from IDENT server.");
		return -1;
	}
	info(VERBOSE, "IDENT resolved remote user to %s.", user);
	safe_strcpy(remote_user, user);
	#ifndef NOFORK
	if (topt & T_SETREMOTE_USER)
		anubis_changeowner(remote_user);
	#endif /* not NOFORK */
	if (topt & T_AUTH_LOCAL)
		tunnel_auth();
	return 0;
}

static void tunnel_auth(void)
{
	info(NORMAL, "Local auth. is ON. Remote user is %s.", remote_user);
	open_config_file(REMOTE, remote_user);
	#ifndef NOFORK
	read_config_file();
	#endif /* not NOFORK */
	#if !defined(NOREGEX) || defined(PCRE)
	regex_position = get_position(BEGIN_REGEX);
	#endif /* not NOREGEX or PCRE */
	topt &= ~T_AUTH_LOCAL; /* We don't need it anymore. */
	return;
}
#endif /* not NOAUTH */

/* EOF */

